################################################################################
#
#   Approval control module for OpenERP
#   Copyright (C) 2009, Almacom (Thailand) Ltd.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

from osv import osv,fields
import time
import netsvc

class approval_request_rule(osv.osv):
    _name="approval.request.rule"
    _columns={
        "name": fields.char("Name",size=64,required=True),
        "type": fields.selection([("pr","Purchase Requisition"),("po","Purchase Order"),("pv","Payment Voucher")],"Type",required=True),
        "user_id": fields.many2one("res.users","Authorizer",required=True),
        "department_id": fields.many2one("hr.department","Department"),
        "min_amount": fields.float("Minimum Amount"),
        "max_amount": fields.float("Maximum Amount"),
        "min_approvals": fields.integer("Number Of Previous Approvals"),
        "approved_by": fields.many2many("res.users","approval_request_rule_users","rule_id","user_id","Previous Approvers"),
    }
    _order="department_id,type"
    _defaults={
        "name": lambda self,cr,uid,context:self.pool.get("ir.sequence").get(cr,uid,"approval.request.rule"),
    }

    def _check_dept(self,cr,uid,ids):
        for rule in self.browse(cr,uid,ids):
            if rule.type!="pr" and rule.department_id:
                return False
        return True

    _constraints = [
        (_check_dept,'Can only set department for PR control',['department_id']),
    ]
approval_request_rule()

class approval_decision_rule(osv.osv):
    _name="approval.decision.rule"
    _columns={
        "type": fields.selection([("pr","Purchase Requisition"),("po","Purchase Order"),("pv","Payment Voucher")],"Type",required=True),
        "action": fields.selection([("approve","Approve"),("disapprove","Disapprove")],"Action",required=True),
        "department_id": fields.many2one("hr.department","Department"),
        "sequence": fields.integer("Priority",required=True),
        "max_amount": fields.float("Maximum Amount"),
        "min_users": fields.integer("Number Of Required Authorizers",required=True),
        "users": fields.many2many("res.users","approval_decision_rule_users","rule_id","user_id","Authorizers"),
        "name": fields.char("Name",size=64,required=True),
    }
    _order="department_id,type,sequence"
    _defaults={
        "action": lambda *a: "approve",
        "sequence": lambda *a: 1,
        "min_users": lambda *a: 1,
        "name": lambda self,cr,uid,context:self.pool.get("ir.sequence").get(cr,uid,"approval.decision.rule"),
    }

    def _check_dept(self,cr,uid,ids):
        for rule in self.browse(cr,uid,ids):
            if rule.type!="pr" and rule.department_id:
                return False
        return True

    _constraints = [
        (_check_dept,'Can only set department for PR control',['department_id']),
    ]
approval_decision_rule()

class approvable(osv.osv):
    _name="approvable"

    def name_get(self,cr,uid,ids,context=None):
        return [(id,"") for id in ids]

    def _obj_id(self,cr,uid,ids,name,arg,context={}):
        vals={}
        for approv in self.browse(cr,uid,ids):
            res=self.pool.get(approv.approv_model).search(cr,uid,[("approvable_id","=",approv.id)])
            obj_id=res[0]
            vals[approv.id]=obj_id
        return vals

    def _can_approve(self,cr,uid,ids,name,arg,context={}):
        vals={}
        for approv in self.browse(cr,uid,ids):
            if approv.approval_decision_rule_id:
                res=False
            elif uid==1 and approv.approval_requests:
                res=True
            else:
                req_ids=[req.user_id.id for req in approv.approval_requests]
                dec_ids=[dec.user_id.id for dec in approv.approval_decisions]
                res=uid in req_ids and not uid in dec_ids
            vals[approv.id]=res
        return vals

    _columns={
        "approv_model": fields.char("Approvable Model",size=64),
        "approval_requests": fields.one2many("approval.request","approvable_id","Approval Requests",readonly=True),
        "approval_decisions": fields.one2many("approval.decision","approvable_id","Approval Decisions",readonly=True),
        "approval_decision_rule_id": fields.many2one("approval.decision.rule","Approval Decision Rule",readonly=True),
        "approved": fields.boolean("Approved",readonly=True),
        "obj_id": fields.function(_obj_id,method=True,type="integer",string="Object ID"),
        "can_approve": fields.function(_can_approve,method=True,type="boolean",string="Can Approve"),
        "date_decision": fields.datetime("Date Approved/Disapproved",readonly=True),
    }

    def reset_approval(self,cr,uid,ids):
        vals={
            "approval_requests":[(6,0,[])],
            "approval_decisions":[(6,0,[])],
            "approval_decision_rule_id": None,
            "date_decision": None,
        }
        self.write(cr,uid,ids,vals)

    def get_rule_domain(self,cr,uid,approv):
        rule_types={
            "purchase.requisition": "pr",
            "purchase.order": "po",
            "payment.voucher": "pv",
            "purchase.requisition.line": "pr",
            "purchase.order.line": "po",
            "payment.voucher.line": "pv",
        }
        domain=[("type","=",rule_types[approv.approv_model])]
        if approv.approv_model=="purchase.requisition":
            pr=self.pool.get("purchase.requisition").browse(cr,uid,approv.obj_id)
            domain+=[("department_id","=",pr.department_id.id)]
        elif approv.approv_model=="purchase.requisition.line":
            prl=self.pool.get("purchase.requisition.line").browse(cr,uid,approv.obj_id)
            domain+=[("department_id","=",prl.order_id.department_id.id)]
        return domain

    def get_amount(self,cr,uid,approv):
        obj=self.pool.get(approv.approv_model).browse(cr,uid,approv.obj_id)
        # PR?
        if approv.approv_model=="purchase.order":
            amount=obj.amount_total
        elif approv.approv_model=="payment.voucher":
            amount=obj.amount_total
        elif approv.approv_model=="purchase.order.line":
            amount=obj.order_id.amount_total
        elif approv.approv_model=="payment.voucher.line":
            amount=obj.voucher_id.amount_total
        else:
            amount=None
        return amount

    def get_lines(self,cr,uid,obj,model):
        if model=="purchase.requisition" or model=="payment.voucher":
            return obj.lines
        elif model=="purchase.order":
            return obj.order_line

    def try_approval_request_rule(self,cr,uid,approv,rule):
        if rule.user_id.id in [req.user_id.id for req in approv.approval_requests]:
            return False
        amount=self.get_amount(cr,uid,approv)
        if rule.min_amount and amount<rule.min_amount:
            return False
        if rule.max_amount and amount>rule.max_amount:
            return False
        user_ids=[dec.user_id.id for dec in approv.approval_decisions if dec.action=="approve"]
        num_approvals=0
        for user in rule.approved_by:
            if user.id in user_ids:
                num_approvals+=1
        if num_approvals<rule.min_approvals:
            return False
        return True

    def apply_approval_request_rules(self,cr,uid,ids):
        for approv in self.browse(cr,uid,ids):
            print "applying approval request rules for object %s %d"%(approv.approv_model,approv.obj_id)
            rule_domain=self.get_rule_domain(cr,uid,approv)
            rule_ids=self.pool.get("approval.request.rule").search(cr,uid,rule_domain)
            for rule in self.pool.get("approval.request.rule").browse(cr,uid,rule_ids):
                print "trying rule %s"%rule.name
                if self.try_approval_request_rule(cr,uid,approv,rule):
                    print "MATCH"
                    vals={
                        "approvable_id": approv.id,
                        "user_id": rule.user_id.id,
                        "rule_id": rule.id,
                    }
                    self.pool.get("approval.request").create(cr,uid,vals)
                    if approv.approv_model in ("purchase.requisition","purchase.order","payment.voucher"):
                        obj=self.pool.get(approv.approv_model).browse(cr,uid,approv.obj_id)
                        for line in self.get_lines(cr,uid,obj,approv.approv_model):
                            vals={
                                "approvable_id": line.approvable_id.id,
                                "user_id": rule.user_id.id,
                                "rule_id": rule.id,
                            }
                            self.pool.get("approval.request").create(cr,uid,vals)

    def try_approval_decision_rule(self,cr,uid,approv,rule):
        amount=self.get_amount(cr,uid,approv)
        if rule.max_amount and amount>rule.max_amount:
            return False
        user_ids=[dec.user_id.id for dec in approv.approval_decisions if dec.action==rule.action]
        if 1 in user_ids:
            return True
        num_users=0
        for user in rule.users:
            if user.id in user_ids:
                num_users+=1
        if num_users<rule.min_users:
            return False
        return True

    def apply_approval_decision_rules(self,cr,uid,ids):
        for approv in self.browse(cr,uid,ids):
            print "applying approval decision rules for object %s %d"%(approv.approv_model,approv.obj_id)
            rule_domain=self.get_rule_domain(cr,uid,approv)
            rule_ids=self.pool.get("approval.decision.rule").search(cr,uid,rule_domain)
            for rule in self.pool.get("approval.decision.rule").browse(cr,uid,rule_ids):
                print "trying rule %s"%rule.name
                if self.try_approval_decision_rule(cr,uid,approv,rule):
                    print "MATCH"
                    self.write(cr,uid,approv.id,{"approval_decision_rule_id":rule.id,"approved":rule.action=="approve","date_decision":time.strftime("%Y-%m-%d %H:%M:%S")})
                    if approv.approv_model in ("purchase.requisition","purchase.order","payment.voucher"):
                        wfs=netsvc.LocalService("workflow")
                        wfs.trg_validate(uid,approv.approv_model,approv.obj_id,rule.action=="approve" and "approved" or "disapproved",cr)
                        break

    def add_decision(self,cr,uid,id,action,reason=None,disapproved_lines={}):
        approv=self.browse(cr,uid,id)
        if not approv.can_approve:
            raise osv.except_osv(_('Error !'), _('You are not allowed to make approval decision for this object'))
        vals={
            "approvable_id": approv.id,
            "user_id": uid,
            "action": action,
            "reason": reason,
        }
        self.pool.get("approval.decision").create(cr,uid,vals)
        self.apply_approval_decision_rules(cr,uid,[approv.id])
        if approv.approv_model in ("purchase.requisition","purchase.order","payment.voucher"):
            self.apply_approval_request_rules(cr,uid,[approv.id])
            obj=self.pool.get(approv.approv_model).browse(cr,uid,approv.obj_id)
            if action=="disapprove":
                for line in self.get_lines(cr,uid,obj,approv.approv_model):
                    self.add_decision(cr,uid,line.approvable_id.id,"disapprove",reason)
            elif action=="approve":
                for line in self.get_lines(cr,uid,obj,approv.approv_model):
                        if line.id in disapproved_lines:
                            reason=disapproved_lines[line.id]
                            self.add_decision(cr,uid,line.approvable_id.id,"disapprove",reason)
                        else:
                            self.add_decision(cr,uid,line.approvable_id.id,"approve")
approvable()

class approval_request(osv.osv):
    _name="approval.request"
    _columns={
        "approvable_id": fields.many2one("approvable","Approvable",required=True),
        "user_id": fields.many2one("res.users","User",required=True),
        "date": fields.datetime("Date Submitted",required=True),
        "rule_id": fields.many2one("approval.request.rule","Rule",required=True),
    }
    _defaults={
        "date": lambda *a: time.strftime("%Y-%m-%d %H:%M:%S"),
    }
approval_request()

class approval_decision(osv.osv):
    _name="approval.decision"
    _columns={
        "approvable_id": fields.many2one("approvable","Approvable",required=True),
        "user_id": fields.many2one("res.users","User",required=True),
        "date": fields.datetime("Date",required=True),
        "action": fields.selection([("approve","Approve"),("disapprove","Disapprove")],"Action",required=True),
        "reason": fields.char("Reason",size=128),
    }
    _defaults={
        "date": lambda *a: time.strftime("%Y-%m-%d %H:%M:%S"),
    }
approval_decision()

class purchase_requisition(osv.osv):
    _name="purchase.requisition"
    _inherit="purchase.requisition"
    _inherits={"approvable": "approvable_id"}

    _columns={
        "approvable_id": fields.many2one("approvable","Approvable",required=True),
    }
    _defaults={
        "approv_model": lambda *a: "purchase.requisition",
    }

    def wkf_draft(self,cr,uid,ids,context={}):
        super(purchase_requisition,self).wkf_draft(cr,uid,ids,context)
        for pr in self.browse(cr,uid,ids):
            self.pool.get("approvable").reset_approval(cr,uid,[pr.approvable_id.id])
            for line in pr.lines:
                self.pool.get("approvable").reset_approval(cr,uid,[line.approvable_id.id])
        return True

    def wkf_submitted(self,cr,uid,ids,context={}):
        super(purchase_requisition,self).wkf_submitted(cr,uid,ids,context)
        for pr in self.browse(cr,uid,ids):
            self.pool.get("approvable").apply_approval_request_rules(cr,uid,[pr.approvable_id.id])
        return True
purchase_requisition()

class purchase_requisition_line(osv.osv):
    _name="purchase.requisition.line"
    _inherit="purchase.requisition.line"
    _inherits={"approvable": "approvable_id"}

    _columns={
        "approvable_id": fields.many2one("approvable","Approvable",required=True),
    }
    _defaults={
        "approv_model": lambda *a: "purchase.requisition.line",
    }
purchase_requisition_line()

class purchase_order(osv.osv):
    _name="purchase.order"
    _inherit="purchase.order"
    _inherits={"approvable": "approvable_id"}

    _columns={
        "approvable_id": fields.many2one("approvable","Approvable",required=True),
        'state': fields.selection([('draft', 'Request for Quotation'), ('wait', 'Submitted For Approval'), ('confirmed', 'Approved'), ('approved', 'Confirmed By Supplier'),('except_picking', 'Shipping Exception'), ('except_invoice', 'Invoice Exception'), ('done', 'Done'), ('cancel', 'Cancelled'),("disapproved","Disapproved")], 'Order Status', readonly=True, help="The state of the purchase order or the quotation request. A quotation is a purchase order in a 'Draft' state. When the order is ready to be submitted for internal approval, the state changes to 'Submitted For Approval'. When the order is approved, the state becomes 'Approved'. Then the supplier must confirm the order to change the state to 'Confirmed By Supplier'. When the purchase order is paid and received, the state becomes 'Done'. If a cancel action occurs in the invoice or in the reception of goods, the state becomes in exception.", select=True),
        "submitted_by": fields.many2one("res.users","Submitted For Approval By",readonly=True),
        "date_submitted": fields.datetime("Date Submitted",readonly=True),
    }
    _defaults={
        "approv_model": lambda *a: "purchase.order",
    }

    def wkf_submitted(self,cr,uid,ids,context={}):
        date=time.strftime("%Y-%m-%d %H:%M:%S")
        self.write(cr,uid,ids,{"state":"wait","date_submitted":date,"submitted_by":uid})
        for po in self.browse(cr,uid,ids):
            self.pool.get("approvable").apply_approval_request_rules(cr,uid,[po.approvable_id.id])
        return True

    def wkf_disapproved(self,cr,uid,ids,context={}):
        self.write(cr,uid,ids,{"state":"disapproved"})
purchase_order()

class purchase_order_line(osv.osv):
    _name="purchase.order.line"
    _inherit="purchase.order.line"
    _inherits={"approvable": "approvable_id"}

    _columns={
        "approvable_id": fields.many2one("approvable","Approvable",required=True),
    }
    _defaults={
        "approv_model": lambda *a: "purchase.order.line",
    }
purchase_order_line()

class payment_voucher(osv.osv):
    _name="payment.voucher"
    _inherit="payment.voucher"
    _inherits={"approvable": "approvable_id"}

    _columns={
        "approvable_id": fields.many2one("approvable","Approvable",required=True),
    }
    _defaults={
        "approv_model": lambda *a: "payment.voucher",
    }

    def wkf_draft(self,cr,uid,ids,context={}):
        super(payment_voucher,self).wkf_draft(cr,uid,ids,context)
        for pv in self.browse(cr,uid,ids):
            self.pool.get("approvable").reset_approval(cr,uid,[pv.approvable_id.id])
            for line in pv.lines:
                self.pool.get("approvable").reset_approval(cr,uid,[line.approvable_id.id])
        return True

    def wkf_submitted(self,cr,uid,ids,context={}):
        super(payment_voucher,self).wkf_submitted(cr,uid,ids,context)
        for pv in self.browse(cr,uid,ids):
            self.pool.get("approvable").apply_approval_request_rules(cr,uid,[pv.approvable_id.id])
        return True
payment_voucher()

class payment_voucher_line(osv.osv):
    _name="payment.voucher.line"
    _inherit="payment.voucher.line"
    _inherits={"approvable": "approvable_id"}

    _columns={
        "approvable_id": fields.many2one("approvable","Approvable",required=True),
    }
    _defaults={
        "approv_model": lambda *a: "payment.voucher.line",
    }
payment_voucher_line()
